//
//  SourceInfoViewController.swift
//  Aidoku (iOS)
//
//  Created by Skitty on 2/15/22.
//

import UIKit
import SafariServices

class SourceInfoViewController: SettingsTableViewController {

    var source: Source

    let languageCodes = [
        "en": "English",
        "ja": "Japanese",
        "zh": "Chinese (Simplified)",
        "zh-hk": "Chinese (Traditional)",
        "es": "Spanish",
        "es-la": "Spanish (Latin America)",
        "fr": "French",
        "it": "Italian",
        "de": "German",
        "nl": "Dutch",
        "pl": "Polish",
        "ru": "Russian",
        "hu": "Hungarian",
        "fi": "Finnish",
        "vi": "Vietnamese",
        "el": "Greek",
        "bg": "Bulgarian",
        "pt-br": "Portuguese (Brazil)",
        "pt": "Portuguese (Portugal)",
        "sv": "Swedish",
        "no": "Norwegian",
        "ar": "Arabic",
        "da": "Danish",
        "bn": "Bengali",
        "ro": "Romanian",
        "cs": "Czech",
        "mn": "Mongolian",
        "tr": "Turkish",
        "id": "Indonesian",
        "ko": "Korean",
        "fa": "Persian",
        "ms": "Malay",
        "th": "Thai",
        "ca": "Catalan",
        "tl": "Filipino",
        "uk": "Ukranian",
        "my": "Burmese",
        "lt": "Lithuanian",
        "he": "Hebrew",
        "hi": "Hindi",
        "ne": "Nepali",
        "sh": "Serbo-Croatian",
        "other": "Other"
    ]

    init(source: Source) {
        self.source = source
        super.init(items: source.settingItems)
    }

    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override func viewDidLoad() {
        super.viewDidLoad()

        title = NSLocalizedString("SOURCE_INFO", comment: "")
        navigationController?.navigationBar.prefersLargeTitles = true

        navigationItem.rightBarButtonItem = UIBarButtonItem(
            barButtonSystemItem: .done,
            target: self,
            action: #selector(close)
        )

        if #available(iOS 15.0, *) {
            tableView.sectionHeaderTopPadding = 0
        }
        tableView.delaysContentTouches = false
        tableView.keyboardDismissMode = .onDrag

        let headerView = SourceInfoHeaderView(source: source)
        headerView.frame.size.height = 48 + 32 + 20
        tableView.tableHeaderView = headerView

        tableView.register(UITableViewCell.self, forCellReuseIdentifier: "UITableViewCell")
        tableView.register(TextInputTableViewCell.self, forCellReuseIdentifier: "TextInputTableViewCell")
    }

    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        tableView.reloadData()
    }

    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        source.needsFilterRefresh = true
    }

    @objc func close() {
        dismiss(animated: true)
    }
}

// MARK: - Table View Data Source
extension SourceInfoViewController {

    override func numberOfSections(in tableView: UITableView) -> Int {
        1 + source.settingItems.count + (source.languages.isEmpty ? 0 : 1)
    }

    override func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        if !source.languages.isEmpty && section == 0 {
            return 1
        } else if source.settingItems.isEmpty || section == source.settingItems.count + (source.languages.isEmpty ? 0 : 1) {
            return 2
        }
        return source.settingItems[section + (source.languages.isEmpty ? 0 : -1)].items?.count ?? 0
    }

    override func tableView(_ tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        if !source.languages.isEmpty && section == 0 {
            return NSLocalizedString("LANGUAGE", comment: "")
        } else if source.settingItems.isEmpty || section == source.settingItems.count + (source.languages.isEmpty ? 0 : 1) {
            return NSLocalizedString("INFO", comment: "")
        }
        return source.settingItems[section + (source.languages.isEmpty ? 0 : -1)].title
    }

    override func tableView(_ tableView: UITableView, titleForFooterInSection section: Int) -> String? {
        if !source.languages.isEmpty && section == 0 {
            return nil
        } else if source.settingItems.isEmpty || section == source.settingItems.count + (source.languages.isEmpty ? 0 : 1) {
            return nil
        }
        return source.settingItems[section + (source.languages.isEmpty ? 0 : -1)].footer
    }

    override func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        if !source.languages.isEmpty && indexPath.section == 0 { // Language selection
            var cell: UITableViewCell? = tableView.dequeueReusableCell(withIdentifier: "UITableViewCell.Value1")
            if cell == nil {
                cell = UITableViewCell(style: .value1, reuseIdentifier: "UITableViewCell.Value1")
            }

            cell?.textLabel?.text = NSLocalizedString("LANGUAGE", comment: "")
//            if let value = UserDefaults.standard.array(forKey: "\(source.id).languages").first,
//               let index = source.languages.firstIndex(of: value) {
//                cell?.detailTextLabel?.text = source.languages[index]
//            }
            cell?.detailTextLabel?.text = nil
            cell?.accessoryType = .disclosureIndicator

            return cell!
        } else if source.settingItems.isEmpty
                    || indexPath.section == source.settingItems.count + (source.languages.isEmpty ? 0 : 1) { // Footer info
            var cell: UITableViewCell? = tableView.dequeueReusableCell(withIdentifier: "UITableViewCell.Value1")
            if cell == nil {
                cell = UITableViewCell(style: .value1, reuseIdentifier: "UITableViewCell.Value1")
            }

            switch indexPath.row {
            case 0:
                cell?.textLabel?.text = NSLocalizedString("VERSION", comment: "")
                cell?.detailTextLabel?.text = String(source.manifest.info.version)
            case 1:
                cell?.textLabel?.text = NSLocalizedString("LANGUAGE", comment: "")
                cell?.detailTextLabel?.text = source.manifest.info.lang
            default:
                break
            }

            cell?.accessoryType = .none
            cell?.selectionStyle = .none

            return cell!
        } else if let item = source.settingItems[indexPath.section + (source.languages.isEmpty ? 0 : -1)].items?[indexPath.row] { // Settings
            let cell = self.tableView(tableView, cellForRowAt: indexPath, settingItem: item)
            if item.type == "text" || item.type == "segment" {
                (cell as? TextInputTableViewCell)?.source = source
            }
            return cell
        }

        return UITableViewCell()
    }

    override func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        if !source.languages.isEmpty && indexPath.section == 0 {
            let item = SettingItem(
                type: "multi-select",
                key: "\(source.id).languages",
                title: NSLocalizedString("LANGUAGE", comment: ""),
                values: source.languages.map { $0.value ?? $0.code },
                titles: source.languages.map { languageCodes[$0.code] ?? $0.code },
                notification: "languageChange"
            )
            navigationController?.pushViewController(SettingSelectViewController(source: source, item: item, style: tableView.style), animated: true)
        } else if source.settingItems.isEmpty || indexPath.section == source.settingItems.count + (source.languages.isEmpty ? 0 : 1) {
            // info
        } else if let item = source.settingItems[indexPath.section + (source.languages.isEmpty ? 0 : -1)].items?[indexPath.row] {
            if item.type == "select" || item.type == "multi-select" {
                navigationController?.pushViewController(
                    SettingSelectViewController(source: source, item: item, style: tableView.style),
                    animated: true
                )
            } else if item.type == "button" {
                if let key = item.action {
                    source.performAction(key: key)
                }
            } else {
                performAction(for: item)
            }
        }

        tableView.deselectRow(at: indexPath, animated: true)
    }
}
